/*****************************************************************************
 * Go - 3D Graphics Library, written by Per Gullberg
 * Version: 1.0
 * Copyright (C) 2000  Hewlett-Packard Company
 * 
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 * 
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *****************************************************************************/


package go;

import java.awt.*;
import java.awt.image.*;
import java.awt.event.*;

class EventQ
{
    int     type;
    EventQ  next;    
    int     x, y;
    char    key;

    EventQ()
    {
	next = null;
    }
}

public abstract class GoInterface extends Canvas implements Runnable
{
    EventQ head = new EventQ();
    EventQ tail = head;

    final static int RENDER         = 1;
    final static int RESIZED        = 2;
    final static int KEY_TYPED      = 3;
    final static int MOUSE_PRESSED  = 4;
    final static int MOUSE_RELEASED = 5;
    final static int MOUSE_DRAGGED  = 6;

    /** Variable containing the Go context that was created with this interface.
     */
    public Go  go;

    //
    // Java image variables.
    //
    MemoryImageSource source;
    Image             image = null;
    int               width = 1;
    int               height = 1;
    boolean           dirty = true;

    /** Constructor for creating a new Go rendering context.

        <P>This will create a new Go context which can be accessed via the variable  <A HREF="GoInterface.html#go">go</A>.
     */
    public GoInterface()
    {
        //
        // Greate a new Go context.
        //
        initializeGo();

	//
	// Resized event listener.
	//
        addComponentListener
	    (
	     new ComponentAdapter()
		 {
		     public void componentResized(ComponentEvent event)
		     {
			 width  = getSize().width;
			 height = getSize().height;

			 //
			 // Resize go.
			 //
			 go.size(width, height);
                
			 //
			 // Resize the Java image.
			 //
			 source = new MemoryImageSource(width, height,
							go.image.buf, 0, width);
			 source.setAnimated(true);
			 image = createImage(source);

			 tail = tail.next = new EventQ();
			 tail.type = RESIZED;
			 tail.x = width;
			 tail.y = height;

			 notifyRenderThread();
		     }
		 }
	     );

	//
	// Keyboard event listener.
	//
	addKeyListener
            (
	     new KeyAdapter()
		 {
		     public synchronized void keyTyped(KeyEvent event)
		     {
			 tail = tail.next = new EventQ();
			 tail.type = KEY_TYPED;
			 tail.key = event.getKeyChar();

			 notifyRenderThread();
		     }
		 }
	     );

	//
	// Mouse pressed and released event listener.
	//
	addMouseListener
            (
	     new MouseAdapter()
		 {
		     public synchronized void mousePressed(MouseEvent event)
		     {
			 tail = tail.next = new EventQ();
			 tail.type = MOUSE_PRESSED;
			 tail.x = event.getX();
			 tail.y = event.getY();

			 notifyRenderThread();
		     }
                
		     public synchronized void mouseReleased(MouseEvent event)
		     {
			 tail = tail.next = new EventQ();
			 tail.type = MOUSE_RELEASED;
			 tail.x = event.getX();
			 tail.y = event.getY();

			 notifyRenderThread();
		     }
		 }
	     );
        
	//
	// Mouse dragged event listener.
	//
	addMouseMotionListener
            (
	     new MouseMotionAdapter()
		 {
		     public synchronized void mouseDragged(MouseEvent event)
		     {
			 tail = tail.next = new EventQ();
			 tail.type = MOUSE_DRAGGED;
			 tail.x = event.getX();
			 tail.y = event.getY();

			 notifyRenderThread();
		     }
		 }
	     );

	//
	// Create and start the render thread.
	//
	Thread t = new Thread(this);
	t.start();
    }

    /** Will initialize/reinitialize Go.
     */
    public void initializeGo()
    {
	//
        // Greate a new Go context.
        //
        go = new Go(width, height);

        //
        // Create the Java image.
        //
        source = new MemoryImageSource(width, height,
                                       go.image.buf, 0, width);
        source.setAnimated(true);
        image = createImage(source);
    }
    
    /** Will force <A HREF="GoInterface.html#render()">render</A>( ) to be called.
     */
    public void rerender()
    {
        dirty = true;
        repaint();
    }

    /** Main rendering method.

        <P>This is the rendering thread callback method for displaying graphics.  This method is triggered by calling <A HREF="GoInterface.html#rerender()">rerender</A>( ).

        <P>Override this method with your rendering code.
     */
    public void render()
    {
	// overridden
    }

    /** Context resized method.

	<P>This is the rendering thread callback method for delivering new context size information.  This method is automatically triggered when a resize event occurs.

	<P>The <FONT COLOR="#0000FF"><TT>width</TT></FONT> and <FONT COLOR="#0000FF"><TT>height</TT></FONT> arguments indicate the new rendering context size in pixels.

        <P>Override this method with your resize specific code.
     */
    public void resized(int width, int height)
    {
	// overridden
    }

    /** Keyboard input method.

        <P>This is the rendering thread callback method for delivering keyboard input information.  This method is automatically triggered when a key is typed on the keyboard.

	<P>The <FONT COLOR="#0000FF"><TT>key</TT></FONT> argument is the ASCII character code of the key typed.

        <P>Override this method with your keyboard input specific code.
     */
    public void keyTyped(char key)
    {
	// overridden
    }

    /** Mouse pressed method.

        <P>This is the rendering thread callback method for delivering  mouse pressed information.  This method is automatically triggered when a mouse button has been pressed.

	<P>The (<FONT COLOR="#0000FF"><TT>x</TT></FONT>, <FONT COLOR="#0000FF"><TT>y</TT></FONT>) arguments indicate the pixel coordinate at which the mouse was pressed.

        <P>Override this method with your mouse pressed specific code.
     */
    public void mousePressed(int x, int y)
    {
	// overridden
    }

    /** Mouse released method.

        <P>This is the rendering thread callback method for delivering  mouse released information.  This method is automatically triggered when a mouse button has been released.

	<P>The (<FONT COLOR="#0000FF"><TT>x</TT></FONT>, <FONT COLOR="#0000FF"><TT>y</TT></FONT>) arguments indicate the pixel coordinate at which the mouse was released.

        <P>Override this method with your mouse released specific code.
     */
    public void mouseReleased(int x, int y)
    {
	// overridden
    }

    /** Mouse dragged method.

        <P>This is the rendering thread callback method for delivering  mouse dragged information.  This method is automatically triggered when a mouse button has been pressed and the mouse is being dragged.

	<P>The (<FONT COLOR="#0000FF"><TT>x</TT></FONT>, <FONT COLOR="#0000FF"><TT>y</TT></FONT>) arguments indicate the current pixel coordinate of the mouse.

        <P>Override this method with your mouse dragged specific code.
     */
    public void mouseDragged(int x, int y)
    {
	// overridden
    }

    /** All rendering is done off screen - calling this method will display the current off screen rendering buffer.
     */
    public void swap()
    {
        Graphics g = getGraphics();

        if(g != null) {

            //
            // Update the Java image.
            //
            source.newPixels(0, 0, width, height);

            //
            // Display the Java image.
            //
            g.drawImage(image, 0, 0, null);
        }
    }

    //
    // Main render loop.
    //
    public synchronized void run()
    {
	while(true) {
	    
	    //
	    // Go to sleep if there are no events to process.
	    //

	    try {
		while(head.next == null) {
		    wait();
		}
	    }
	    catch (InterruptedException e) {
		System.out.println("ERROR: event handler malfunction");
	    }


	    //
	    // Not sleeping, and ready to process events.
	    //

	    EventQ event = head.next;
	    
	    while(event != null) {
	    
		switch(event.type) {
	
		case RENDER:
		    render();
		    break;

		case RESIZED:
		    resized(event.x, event.y);
		    rerender();
		    break;

		case KEY_TYPED:
		    keyTyped(event.key);
		    break;

		case MOUSE_PRESSED:
		    mousePressed(event.x, event.y);
		    break;

		case MOUSE_RELEASED:
		    mouseReleased(event.x, event.y);
		    break;

		case MOUSE_DRAGGED:
		    mouseDragged(event.x, event.y);
		    break;
		}

		event = event.next;
	    }
	    
	    head.next = null;
	    tail = head;
	}
    }

    //
    // Thread notification that an event has arrived.
    //
    synchronized void notifyRenderThread()
    {
	notifyAll();
    }

    //
    // Overrides update in class java.awt.Component
    //
    public synchronized void update(Graphics g)
    {
        paint(g);
    }

    //
    // Overrides paint in class java.awt.Canvas
    //
    public synchronized void paint(Graphics g)
    {
        //
        // If dirty, then rerender with Go.
        //
        if(dirty) {
            dirty = false;
            tail = tail.next = new EventQ();
	    tail.type = RENDER;
	    notifyRenderThread();
        }

        //
        // Display the Java image.
        //
        g.drawImage(image, 0, 0, null);
    }
}
